@@ -19,26 +19,44 @@
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
+/* User bookmarks come from the file ~/.gtk-bookmarks.
+ *
+ * When changed() is first called, it returns TRUE.
+ *
+ * When get_bookmarks() is first called, they will be built using
+ * pbuser_build_bookmarks(). It stores the bookmarks in PBUserData.bookmarks
+ * and the file's mtime in PBUserData.loaded. get_bookmarks() then clones the
+ * bookmarks from PBUserData.bookmarks.
+ *
+ * Once that's done, a call to changed() checks the file's mtime and a couple
+ * other things.
+ */
+
 #ifdef HAVE_CONFIG_H
 #  include <config.h>
 #endif
 
 #include "model_user.h"
 #include "model.h"
+#include "support.h"
+
+#include <sys/types.h>
+#include <sys/stat.h>
 
 #include <libxfce4util/libxfce4util.h>
 #include <glib.h>
 #include <glib/gstdio.h>
 
 #define pbg_priv(pbg) ((PBUserData*) pbg->priv)
-#define show_bookmark(b) ((gboolean) b->priv)
-
+#define show_bookmark(b) (GPOINTER_TO_INT(b->priv))
 
 typedef struct
 {
   GList     *bookmarks;
   gchar     *filename;
-  time_t     loaded;
+  time_t     loaded; /* 0  indicates loading the file hasn't been attempted
+                        1  indicates the file does not exist
+                        2+ are actual timestamps */
 
 } PBUserData;
 
@@ -47,8 +65,9 @@ pbuser_get_mtime(const gchar *filename)
 {
     struct stat buf;
     if(g_stat(filename, &buf) == 0)
-        return buf.st_mtime;
-    return 0;
+        return MAX(buf.st_mtime, 2);
+    else
+        return 1;
 }
 
 static inline gboolean
@@ -58,12 +77,18 @@ pbuser_dir_exists(const gchar *path)
 }
 
 static void
-pbuser_free_bookmark(PlacesBookmark *bookmark)
+pbuser_finalize_bookmark(PlacesBookmark *bookmark)
 {
     g_assert(bookmark != NULL);
-    g_free(bookmark->uri);
-    g_free(bookmark->label);
-    g_free(bookmark);
+
+    if(bookmark->uri != NULL){
+        g_free(bookmark->uri);
+        bookmark->uri = NULL;
+    }
+    if(bookmark->label != NULL){
+        g_free(bookmark->label);
+        bookmark->label = NULL;
+    }
 }
 
 static void
@@ -74,12 +99,16 @@ pbuser_destroy_bookmarks(PlacesBookmarkGroup *bookmark_group)
     if(bookmarks == NULL)
         return;
     
+    DBG("destroy internal bookmarks");
+
     while(bookmarks != NULL){
-        pbuser_free_bookmark((PlacesBookmark*) bookmarks->data);
+        places_bookmark_destroy((PlacesBookmark*) bookmarks->data);
         bookmarks = bookmarks->next;
     }
     g_list_free(bookmarks);
     pbg_priv(bookmark_group)->bookmarks = NULL;
+
+    pbg_priv(bookmark_group)->loaded = 0;
 }
 
 static void
@@ -100,6 +129,7 @@ pbuser_build_bookmarks(PlacesBookmarkGroup *bookmark_group)
 
     if(G_UNLIKELY(fp == NULL)){
         DBG("Error opening gtk bookmarks file");
+        pbg_priv(bookmark_group)->loaded = 1;
         return;
     }
 
@@ -136,12 +166,11 @@ pbuser_build_bookmarks(PlacesBookmarkGroup *bookmark_group)
         }
 
         /* create the BookmarkInfo container */
-        bookmark        = g_new0(PlacesBookmark, 1);
-        bookmark->uri   = path;                                   /* needs to be freed */
-        bookmark->label = name;                                   /* needs to be freed */
-        bookmark->icon  = "gnome-fs-directory";
-        bookmark->priv  = (gpointer) pbuser_dir_exists(path);
-        bookmark->free  = pbuser_free_bookmark;
+        bookmark        = places_bookmark_create(name);           /* label needs to be freed */
+        bookmark->uri   = path;                                   /* uri   needs to be freed */
+        bookmark->icon  = "folder";
+        bookmark->priv  = GINT_TO_POINTER(pbuser_dir_exists(path));
+        bookmark->finalize = pbuser_finalize_bookmark;
 
         bookmarks = g_list_prepend(bookmarks, bookmark);
     }
@@ -156,39 +185,62 @@ pbuser_build_bookmarks(PlacesBookmarkGroup *bookmark_group)
 static GList*
 pbuser_get_bookmarks(PlacesBookmarkGroup *bookmark_group)
 {
-    GList *bookmarks    = pbg_priv(bookmark_group)->bookmarks;
-    GList *clone        = NULL;
-    PlacesBookmark *bookmark;
+    const GList *orig_ls      = pbg_priv(bookmark_group)->bookmarks;
+    const PlacesBookmark *orig;
+    
+    GList *clone_ls           = NULL;
+    PlacesBookmark *clone;
 
-    if(bookmarks == NULL){
+    PlacesBookmarkAction *open, *terminal;
+
+
+    if(orig_ls == NULL){
         pbuser_build_bookmarks(bookmark_group);
-        bookmarks = pbg_priv(bookmark_group)->bookmarks;
+        orig_ls = pbg_priv(bookmark_group)->bookmarks;
+
+        if(orig_ls == NULL)
+            return NULL;
     }
 
-    bookmarks = g_list_last(bookmarks);
+    orig_ls = g_list_last((GList*) orig_ls);
 
-    while(bookmarks != NULL){
-        bookmark        = g_memdup(bookmarks->data, sizeof(PlacesBookmark));
-        bookmark->uri   = g_strdup(bookmark->uri);
-        bookmark->label = g_strdup(bookmark->label);
+    while(orig_ls != NULL){
 
-        clone = g_list_prepend(clone, bookmark);
-        bookmarks = bookmarks->prev;
+        orig = (PlacesBookmark*) orig_ls->data;
+
+        if(show_bookmark(orig)){
+
+            clone                 = places_bookmark_create(g_strdup(orig->label));
+            clone->uri            = g_strdup(orig->uri);
+            clone->uri_scheme     = orig->uri_scheme;
+            clone->icon           = orig->icon;
+            clone->finalize       = pbuser_finalize_bookmark;
+    
+            terminal              = places_create_open_terminal_action(clone);
+            clone->actions        = g_list_prepend(clone->actions, terminal);
+            open                  = places_create_open_action(clone);
+            clone->actions        = g_list_prepend(clone->actions, open);
+            clone->primary_action = open;
+    
+            clone_ls = g_list_prepend(clone_ls, clone);
+        }
+        orig_ls  = orig_ls->prev;
     }
 
-    return clone;
+    return clone_ls;
 
 }
 
 static gboolean
 pbuser_changed(PlacesBookmarkGroup *bookmark_group)
 {
+    /* If we haven't even tried, we should load the bookmarks */
     if(pbg_priv(bookmark_group)->loaded == 0)
         goto pbuser_did_change;
 
-    /* see if the file has changed */
+    /* see if the file has changed (mtime or existence) */
     time_t mtime = pbuser_get_mtime(pbg_priv(bookmark_group)->filename);
-    if(mtime > pbg_priv(bookmark_group)->loaded)
+    if(mtime != pbg_priv(bookmark_group)->loaded)
         goto pbuser_did_change;
     
     /* see if any directories have been created or removed */
@@ -199,12 +251,16 @@ pbuser_changed(PlacesBookmarkGroup *bookmark_group)
     while(bookmarks != NULL){
         bookmark = bookmarks->data;
         if(show_bookmark(bookmark) != pbuser_dir_exists(bookmark->uri)){
-            bookmark->priv = (gpointer) !show_bookmark(bookmark);
+            bookmark->priv = GINT_TO_POINTER(!show_bookmark(bookmark));
             ret = TRUE;
         }
         bookmarks = bookmarks->next;
     }
-    return ret;
+    if(ret)
+        return TRUE;
+
+    /* if we're still here, assume nothing changed */
+    return FALSE;
 
   pbuser_did_change:
     pbuser_destroy_bookmarks(bookmark_group);
@@ -220,7 +276,7 @@ pbuser_finalize(PlacesBookmarkGroup *bookmark_group)
     pbg_priv(bookmark_group)->filename = NULL;
 
     g_free(pbg_priv(bookmark_group));
-    g_free(bookmark_group);
+    bookmark_group->priv = NULL;
 }
 
 
@@ -229,7 +285,9 @@ pbuser_finalize(PlacesBookmarkGroup *bookmark_group)
 PlacesBookmarkGroup*
 places_bookmarks_user_create()
 { 
-    PlacesBookmarkGroup *bookmark_group = g_new0(PlacesBookmarkGroup, 1);
+    PlacesBookmarkGroup *bookmark_group;
+    
+    bookmark_group = places_bookmark_group_create();
     bookmark_group->get_bookmarks       = pbuser_get_bookmarks;
     bookmark_group->changed             = pbuser_changed;
     bookmark_group->finalize            = pbuser_finalize;
